; [KSC4-ASM]----------------------------------------------------------------[]
; 
;  T12RNG.ASM - Random Number Generator with ATTiny12, version 1.00.000
;               Written by Marco "Kiko" Carnut, 2006-jan-28
;               based on the 2005-oct-19 version for ATTiny11
;
;  Description: This program collects random bits from the analog comparator
;  ------------ (assuming it's hooked up to a whitenoise generator) then
;  performs a van Neumann-style debiasing: bits are paired and two equal bits
;  are discarded; when we get two different bits, the second is fed to a
;  shift register. When we get a whole byte, it is xored with an accumulator.
;  After a certain amount of them are xored, it is then transmitted in one
;  of the output pins as serial data at 300 bps.
;
;  Since we use a busy-waiting software UART, the interbit delay must be
;  calibrated to generate the proper baud rate. In an ATTiny12 with its
;  internal 1.2MHz RC oscillator outputting 300bps it works just fine.
;
;  Portability: Although originally implemented on ATTiny12, it should work
;  ------------ in nearly all AVR devices by simply fixing the busy-waiting
;  constant used for baud rate generation and hooking up the appropriate
;  I/O ports. This program compiled with AvrAssembler2 from Atmel's AVR
;  Studio 4.
;
; []------------------------------------------------------------------------[]

.include "tn12def.inc"

.CSEG
		rjmp	RESET

.def    xoracc    = r16		; Random byte sent (xor of the last xorqty bytes)
.def    xorcnt    = r17		; How many bytes xored so far
.def    rndbyte   = r18     ; Random byte being xored, collect after VN debias
.def    rndbits   = r19     ; How many valid bits in rndbyte so far
.equ	xorqty    = 8		; How many bytes to xor before transmitting

.def	data	  = r20   	; Data to transmit


.equ    txd       = 3     	; Bit in PORTB where the data is transmitted
.def    txbitcnt  = r21     ; How many bits have been shifted so far
.def    txbyte    = r22   	; Byte being transmitted
.equ    udelay    = 143 	; 2400 bps @ 1.2MHz clock from internal RC osc
							; (2395bps actually, -0.2% error). This will be
							; further divided by 8 in the transmission loop,
							; so the serial baud rate is in fact 300bps.
							; Such a slow speed was selected to avoid xmit
							; problems caused by the RC oscillator drift
							; with temperature and voltage
.equ	sb		  = 1		; Number of stop bits (1, 2, ...)

.def	temp1	  = r23     ; Generic counter for delay loops/intermediates
.def	temp2	  = r24     ; Generic counter for intermediate results/delays

.def    flags     = r25     ; Flags and random bits collected
.def    bcnt      = r0      ; How many bytes were output, for CRLF generation

.equ    rndbit1   = 0       ; Random bit before debiasing (must be 0)
.equ    has_bit   = 1       ; rndbit1 is valid
.equ    hi_nibble = 2       ; Transmit high nibble
.equ    lo_nibble = 3       ; Transmit low nibble
.equ    send_cr   = 4       ; Transmit a CR
.equ    send_lf   = 5       ; Transmit a LF

.equ    cr        = 13      ; ASCII code for Carriage Return
.equ    lf        = 10      ; ASCII code for Line Feed
.equ    linelen   = 78      ; wrap lines at this much chars


RESET:
		ldi     temp1,(1<<txd)      ; The 'txd' is set to
		out		ddrb,temp1          ; direction=OUT, all others =IN
		ldi     temp1,~(1<<txd)     ; Enable pullups to
		out     portb,temp1         ; minimize power
		ldi     rndbyte,0           ; character to output
		ldi     rndbits,0
		ldi     xorcnt,0
		ldi     r30,1
		ldi     r31,0
		cbi     ACSR,ACD            ; Enable analog comparator
		sbi		ACSR,AINBG			; Enable bandgap reference

idelay0:							; Inter-bit sampling delay                            	
		ldi     temp1,12
idelay1:
		dec		temp1
		brne    idelay1
		
main:								; The flags indicate we have something
		sbrc    flags,hi_nibble     ; to transmit, do it
		rjmp	prep_hi
		sbrc	flags,lo_nibble
		rjmp	prep_lo
		sbrc    flags,send_cr
		rjmp    prep_cr
		sbrc    flags,send_lf
		rjmp    prep_lf

get_rnd_bit:						; otherwise we sample a random bit

        sbrc    flags,has_bit       ; if it is the second bit, go to the
		rjmp    vn_debias			; von-Neumann debiasing setp

        in      temp1,ACSR          ; otherwise get the output
		;ldi		temp2,0				; ---
		bst     temp1,ACO			; of the analog comparator

		;bld	    temp2,0				; ---
		;rjmp	add_bit				; ---

		bld     flags,rndbit1       ; and store it for later
		sbr     flags,(1<<has_bit)  ; flag that we already have a bit
		rjmp    idelay0             ; go to the main loop

vn_debias:							; sample new bit and do von-Neumann debias
        cbr     flags,(1<<has_bit)  ; clear the 'I have a bit' flag
        ldi     temp2,0
        in      temp1,ACSR			; get the output of the
		bst     temp1,ACO			; analog comparator
		bld     temp2,rndbit1		; and the
		mov     temp1,flags         ; previously stored
		andi    temp1,(1<<rndbit1)  ; bit
		eor     temp1,temp2         ; xor them
		breq    idelay0             ; if equal, discard: go to the main loop

add_bit:                            ; if different, we accumulate it
        ror     temp2               ; send it to the carry (assumes rndbit=0!)
		rol     rndbyte				; rotate left to get it inside the byte 
		inc     rndbits				; increment the number of valid bits
		cpi		rndbits,8           ; if we didn't fill a whole byte yet
		brlo    idelay0             ; go to the main loop
		ldi     rndbits,0           ; otherwise reset the valid bit counter

xor_with_buf:
        ld      xoracc,Z            ; get the value pointed by Z
		eor     xoracc,rndbyte      ; accumulate it in the xor accumulator
		st      Z,xoracc

inc_buf_pointer:
		inc     zl                  ; inc the Z-pointer
		cpi     zl,14
		brlo    idelay0             ; continue if it didn't fill out the buffer
		ldi     zl,1                ; otherwise reset it to the beginning

inc_xor_cnt:
		inc     xorcnt              ; and count how many we accumulated so far
		cpi     xorcnt,xorqty       ; if we didn't have accumulated enough
		brlo    idelay0             ; go to the main loop
	    sbr     flags,(1<<hi_nibble); otherwise, set the 'xmit hi nibble' flag
		ldi     xorcnt,0            ; reset the number of accumulated bytes
		rjmp 	idelay0             ; and go to the main loop

prep_cr:
		ldi		data,CR             ; load a carriage return
		cbr     flags,(1<<send_cr)  ; disable 'send_cr' flag
		sbr     flags,(1<<send_lf)  ; enable  'send_lf' flag for the next pass
		clr     bcnt
		rjmp    xmit                ; transmit

prep_lf:
        ldi     data,LF             ; load a line feed
		cbr     flags,(1<<send_lf)  ; disable 'send_lf' flag
		clr     bcnt
		rjmp    xmit				; transmit

prep_hi:							; prepare a high nibble for transmission
		ld		data,Z
		;mov     data,xoracc         ; get the accumulated byte
		swap    data                ; swap high and low nibbles
		cbr     flags,(1<<hi_nibble); clear the 'xmit high nibble flag'
		sbr     flags,(1<<lo_nibble); set the 'xmit low nibble' for next round
		rjmp    make_hex			; convert it to hex and transmit

prep_lo:							; prepare a low niblle for transmission
        cbr     flags,(1<<lo_nibble); clear the 'xmit low nibble flag'
		;mov     data,xoracc			; get the data, convert to hex and transmit
		ld		data,Z
		inc		zl
		cpi     zl,14
		brlo	prep_lo2
		ldi     zl,1
		rjmp	make_hex
prep_lo2:
		sbr		flags,(1<<hi_nibble)

make_hex:							; convert a nibble to hex digit
		ldi     temp1,15			; clear the
		and		data,temp1          ; high nibble
		ldi		temp1,'0'           ; add ASCII '0' to 
		add		data,temp1          ; put it in '0-9:-?' range
		cpi     data,'9'+1          ; if it's less than 9,
		brlo	xmit                ; its correct so transmit
		ldi     temp1,7             ; otherwise add 7 to put it
		add     data,temp1          ; in the 'A-F' range and trasmit it

xmit:								; software UART: transmit a byte over serial
		mov		txbyte,data         ; get the data to be transmitted

putchar:
		ldi	txbitcnt,9+sb			; 1+8+sb (sb is # of stop bits)
		com	txbyte					; Invert everything
		sec							; Start bit
putchar0:
		brcc	putchar1			; If carry set
		cbi		PORTB,txd			; send a '0'
		rjmp	putchar2			; else	

putchar1:
		sbi		PORTB,txd			; send a '1'
		nop

putchar2:                           ; inter-bit delay
        ldi     temp2,8
UART_delay0:
        ldi		temp1,udelay
UART_delay:
		dec		temp1
		brne	UART_delay
		dec     temp2
		brne    UART_delay0

		ror		txbyte				; Get next bit
		dec	    txbitcnt			; If not all bit sent
		brne	putchar0			; send next
									; else
        inc     bcnt                ; increment num of chars in line sent
        ldi     temp1,linelen		; compare with max line length to see
		cp      bcnt,temp1          ; whether we have already filled up a line
        brlo    go2main             ; if not, go to the main loop
		clr		bcnt				; otherwise, clear the char counter
        sbr     flags,(1<<send_cr)  ; and arrange for a CR to be sent
go2main:
        rjmp    main				; go to the main loop
