/* [KSC4]-------------------------------------------------------------------[]
 *
 *  rngfeed.c - Feeds entropy from a HRNG to the Linux entropy pool,
 *              Version 0.02.002 written by 
 *              Marco "Kiko" Carnut <kiko@tempest.com.br>
 *
 *  Description: 
 *  ------------ 
 *
 *  Licensing: This program is free software released under the GNU General
 *  ---------- Public License Version 2: http://www.gnu.org/copyleft/gpl.html
 *
 *  Portability: Tested under Linux 2.4.29.
 *  ------------ 
 *
 *  Building: A "gcc -o rngfeed rngfeed.c -lm" should be enough.
 *  ---------
 *
 *  Warning: 
 *  -------- 
 *
 *  Revision History:
 *
 *  --REV--- -DATE- ---BY--- --------------------Description------------------
 *  0.03.003 061227 Kiko     Added -P option to monitor progress. Fixed a
 *                           stupid bug that it was saving blocks full of 
 *                           zero bytes. Fixed compilation under Linux.
 *  0.02.002 061221 Kiko     Fixed a few bugs with failure detection.
 *  0.01.001 060701 Kiko     Original version.
 * 
 *
 * []-----------------------------------------------------------------------[]
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdio.h>
#include <string.h>
#include <time.h>
#include <stdlib.h>
#include <math.h>

#ifndef   __CYGWIN__
#include <asm/termios.h>
#else
#include <termios.h>
#endif /* __CYGWIN__ */

/* --------------------------- Macro-declared Constants ------------------- */

#define DEFAULT_INPUT_FILE          "/dev/ttyS0"
#define DEFAULT_POOLSIZE                  10240
#define BAUDRATE                          B1200

#define MAXPOOLS                             16

#define LF                                   10

#define TRUE                                  1
#define FALSE                                 0

#define log2of10         3.32192809488736234787

#ifndef   __CYGWIN__
#define RNDADDENTROPY   _IOW( 'R', 0x03, int [2] )
#endif /* __GYGWIN__ */

/* ---------------------- Declarations and Global Variables --------------- */

char *input_file=DEFAULT_INPUT_FILE;
char *rnd_device=NULL;
char *raw_output_file=NULL;
char *log_file=NULL;

struct our_pool {
    char *pool;
    int poolsize;
    int current;
    struct our_pool *next;
};

struct our_pool *first_pool=NULL,*last_pool=NULL;
int npools=0;

int rnd_fd=-1;
FILE *log_fp; 
FILE *raw_out_fp;  

int save_all_blocks=FALSE;
int show_progress=FALSE;
int bytecount=0;

long poolsize=DEFAULT_POOLSIZE;
char *pool;
int tail;

long ccount[256];
long blkcount;

long naccepteds,nrejecteds;
long nrej_chi,nrej_avg,nrej_scc,nrej_ent;

double scc, sccun, sccu0, sccfirst, scclast, scct1, scct2, scct3;
double ent, chisq, chip, average, aval;

struct _rand_pool_info {
        int     entropy_count;
        int     buf_size;
        char    *buf;
};

static double chsqt0[10] = {
	 0.5,  0.25,   0.1,    0.05,    0.025,
	0.01,  0.005,  0.001,  0.0005,	0.0001
};

static double chsqt1[10] = {
	0.0,    0.6745, 1.2816, 1.6449, 1.9600,
	2.3263, 2.5758, 3.0902, 3.2905, 3.7190
};

/* ---------------------------- Function Prototypes ----------------------- */

void add_to_pool(char rndbyte);
void reset_stats();
void update_stats(int c);
void finalize_stats(long totalc);
void commit();
void usage();
void dump(int nbytes);
void *kalloc(int size, char *file, int line);

/* ------------------------ Macro-defined Functions ---------------------- */

#define NEW(qty,type) ((type *)kalloc(qty*sizeof(type),__FILE__,__LINE__))

/* ------------------------------- Main Program --------------------------- */
       
main(int argc, char **argv)
{
    int serial_fd,res,n,a;
    struct termios oldtio,newtio;
    char waiting_eol,upper_nibble,c;
    int has_set_serial_params=FALSE;
    int val,rndbyte,dashes,dumpcnt;
        
    /* Logs go to standard output by default unless the user says otherwise */    
    log_fp=stdout; 

    /* Parse command line options */
    for (a=0,n=1;n<argc;n++) {
        if (argv[n][0]=='-') {
            switch(argv[n][1]) {
                case 'o': if (argv[n][2]) {
                               raw_output_file=argv[n]+2;
                           } else {
                               raw_output_file=argv[++n];
                           }
                           break;
                case 'l': if (argv[n][2]) {
                               log_file=argv[n]+2;
                           } else {
                               log_file=argv[++n];
                           }
                           break;
                case 'r': if (argv[n][2]) {
                               rnd_device=argv[n]+2;
                           } else {
                               rnd_device=argv[++n];
                           }
                           break;
                case 'p': if (isdigit(argv[n][2])) {
                             poolsize=atoi(argv[n]+2);
                          } else {
                             poolsize=atoi(argv[++n]);
                          }
                          break;
                case 'a': save_all_blocks=TRUE;
                          break;
                case 'h': usage();
                          exit(0);
                case 'P': show_progress=TRUE;
                          break;
            }
        } else {
            switch(a) {
                case 0: input_file=argv[n];
                        break;
            }
            a++;
        }
    }

    if (rnd_device) {
        rnd_fd = open(rnd_device, O_WRONLY);
        if (rnd_fd<0) {
            fprintf(stderr,"Unable to open random input file for writing\n");
            exit(1);
        }
        /* FIXME: Should test whether random device responds to the IOCTLs */
    }
   
    pool=NEW(poolsize,char);
        
    serial_fd = open(input_file, O_RDWR | O_NOCTTY ); 
    if (serial_fd<0) {
        fprintf(stderr,"Unable to open %s\n",input_file); 
        exit(1);
    }
    
    if (!tcgetattr(serial_fd,&oldtio)) {
        bzero(&newtio, sizeof(newtio));
        newtio.c_cflag = BAUDRATE | CS8 | CLOCAL | CREAD | CRTSCTS;
        newtio.c_iflag = IGNPAR;
        newtio.c_oflag = 0;
        cfsetospeed(&newtio,BAUDRATE);
        cfsetispeed(&newtio,BAUDRATE);
        /* set input mode (non-canonical, no echo,...) */
        newtio.c_lflag = 0;
        newtio.c_cc[VTIME]=10;
        newtio.c_cc[VMIN] =1;   
        tcflush(serial_fd, TCIFLUSH);
        tcsetattr(serial_fd,TCSANOW,&newtio);
        has_set_serial_params=TRUE;
        int status;
        ioctl(serial_fd, TIOCMGET, &status);
        status |= TIOCM_DTR | TIOCM_RTS;
        ioctl(serial_fd, TIOCMSET, &status);
    }
    
    if (log_file) {
        if (!(log_fp=fopen(log_file,"wt"))) {
            fprintf(stderr,"Unable to open log file for writing\n");
            exit(1);
        }
    }
    
    if (raw_output_file) {
        if (!(raw_out_fp=fopen(raw_output_file,"w"))) {
            fprintf(stderr,"Unable to open raw output file for writing\n");
            exit(1);        
        }
    }

    reset_stats();
    
    waiting_eol=TRUE;
    upper_nibble=FALSE;
    dashes=0;
    dumpcnt=0;
    for (;;) {      
        res = read(serial_fd,&c,1);
        if (res<1) {
            fprintf(stderr,"Timeout reading from serial. "
                           "Is the HRNG connected and turned on?\n");
            exit(1);
        }
        //fprintf(log_fd,"%c",c);
        if (waiting_eol) {
            if (c==LF) {
                waiting_eol=FALSE;
                upper_nibble=TRUE;
            }
            if (c=='-') {
                if (++dashes>200) {
                    fprintf(stderr,"\n*** Too many dashes -- ");
                    fprintf(stderr,"HRNG malfunction? ***\n");
                    dashes=0;
                }
            }
        } else {
            if (isxdigit(c)) {
                if (c>'Z') c-=32;
                int val=c-'0';
                if (val>9) val-=7;
                if (upper_nibble) {
                    rndbyte=val<<4;
                    upper_nibble=FALSE;
                } else {
                    rndbyte|=val;
                    add_to_pool(rndbyte);
                    //fprintf(log_fd,"[%x]",rndbyte);
                    if (show_progress && !(bytecount%20) && bytecount) {
                        fprintf(stderr,"%5d:",bytecount);
                        unsigned int n;
                        for (n=bytecount-20;n<bytecount;n++) {
                            fprintf(stderr," %02X",(pool[n])&0xFF);
                        }
                        fprintf(stderr,"\r");
                    }
                    bytecount++;
                    // For each two bytes we get, dump some from the pool
                    // to the kernel's RNG
                    if (++dumpcnt==2) {
                        int nbytes=npools/2;
                        if (nbytes<1) nbytes=1;
                        dump(nbytes);
                        dumpcnt=0;
                    }
                    // Reset flag to get upper nibble for the next byte
                    upper_nibble=TRUE;
                }
                dashes=0;
            } else {
                waiting_eol=TRUE;
            }
        }
    }
    if (has_set_serial_params) {
        // Restore serial settings to what they were before
        tcsetattr(serial_fd,TCSANOW,&oldtio);
    }
    if (rnd_fd>=0) close(rnd_fd);
    close(serial_fd);
    free(pool);
} /* main */

void add_to_pool(char rndbyte)
{
    int accepted,committed,all_zeros_block=FALSE;
    /* Append random byte to the pool */
    pool[tail]=rndbyte;
    //fprintf(log_fp,"%d=%02X\n",tail,rndbyte);
    
    update_stats(rndbyte);
    
    /* Is the pool full already? */
    if (++tail>=poolsize) {
        /* Finish stats computations and print/judge results */
        finalize_stats(tail);
        accepted=TRUE;
        committed=FALSE;
        bytecount=0;
        fprintf(log_fp,"%5d ",++blkcount);

        fprintf(log_fp,"%7.3f",average);
        if (average<=126.0 || average>=129.0) {
            accepted=FALSE;
            nrej_avg++;
            fprintf(log_fp,"* ");
        } else {
            fprintf(log_fp,"  ");
        }

        fprintf(log_fp,"%6.4f",ent);
        if (ent<7.95) {
            accepted=FALSE;
            nrej_ent++;
            fprintf(log_fp,"* ");
        } else {
            fprintf(log_fp,"  ");
        }

        fprintf(log_fp,"%7.2f %5.2f%%",chisq,chip*100);
        /* Judge quality */
        if (chip<0.1 || chip>0.9) {
            accepted=FALSE;
            nrej_chi++;
            fprintf(log_fp,"* ");
        } else {
            fprintf(log_fp,"  ");
        }

        fprintf(log_fp,"%+7.5f",scc);
        if (scc<=-0.02 || scc>=0.02) {
            accepted=FALSE;
            nrej_scc++;
            fprintf(log_fp,"* ");
        } else {
            fprintf(log_fp,"  ");
        }
        
        if (raw_out_fp && (accepted || save_all_blocks)) {
            fwrite(pool,1,poolsize,raw_out_fp);   
            fflush(raw_out_fp);
            /* This check for all zeros block was a stupid way to
               check a stupid bug that has long been squashed and thus
               no longer needed. However, since it does no harm, I left it
               there anyway. */
            unsigned int n;
            for (n=0;n<poolsize;n++) {
                if (pool[n]) break;
            }
            if (n>=poolsize) {
                all_zeros_block=TRUE;
            }
        }

        if (accepted) {
            naccepteds++;
            if (rnd_fd>=0) {
                commit(); committed=TRUE;
            }
        } else
            nrejecteds++;

        fprintf(log_fp,"%6d/%-6d",naccepteds,nrejecteds);
        if (committed) {
            fprintf(log_fp," [C]");
        } else {
            fprintf(log_fp,"    ");
        }
        fprintf(log_fp," %02d",npools);
        if (all_zeros_block) fprintf(log_fp," !Z!");
        fprintf(log_fp,"\n");
        /* Reset stats */
        reset_stats();

        /* Reset tail back to the head */
        tail=0;
    }    
}

void commit()
{
    int amount;
    struct our_pool *newpool;
    while (npools>=MAXPOOLS) {
        amount=first_pool->poolsize-first_pool->current;
        dump(amount);
        fprintf(log_fp,"\n-- forced dump of %d bytes, ",amount);
        fprintf(log_fp,"%d pools remaining\n",npools);
    }
    newpool=NEW(1,struct our_pool);
    newpool->poolsize=poolsize;
    newpool->pool=pool;
    newpool->next=NULL;
    newpool->current=0;
    npools++;
    if (last_pool) last_pool->next=newpool;
    last_pool=newpool;
    if (!first_pool) first_pool=newpool;
    pool=NEW(poolsize,char);
}

void dump(int nbytes)
{
    int amount,bytes_left;
    struct _rand_pool_info rpi;

    if (!first_pool) return;
    while (nbytes>0) {
        bytes_left = (first_pool->poolsize) - (first_pool->current);
        amount = nbytes;
        if (amount > bytes_left) amount=bytes_left;
        rpi.entropy_count=amount<<3;
        rpi.buf_size=amount;
        rpi.buf=first_pool->pool+first_pool->current;
#ifndef   __CYGWIN__
        if (rnd_fd>=0) ioctl(rnd_fd, RNDADDENTROPY, &rpi);
#endif /* __CYGWIN__ */
        first_pool->current+=amount;
        if (first_pool->current==first_pool->poolsize) {
            struct our_pool *newfirst=first_pool->next;
            free(first_pool->pool);
            free(first_pool);
            first_pool=newfirst;
            npools--;
        }
        nbytes-=amount;
    }
}

/* ---------------------------- Auxiliary Functions ----------------------- */

void reset_stats()
{
    int i;
    average = 0; aval = 0;
    for (i=0;i<256;i++) ccount[i] = 0;
    sccfirst = 1;	 
    scct1 = scct2 = scct3 = 0.0;
    scc = sccun = sccu0 = scclast = 0;
}

void update_stats(int c)
{
    c&=0xFF;
    aval+=c;                  /* Accumulator for average */
    ccount[c]++;              /* Per-byte counters for entropy */
    
    /* Serial correlation calculation stuff */
    sccun = c;
    if (sccfirst) {
        sccfirst = 0;
        scclast = 0;
        sccu0 = sccun;
    } else {
        scct1 = scct1 + scclast * sccun;
    }
    scct2 = scct2 + sccun;
    scct3 = scct3 + (sccun * sccun);
    scclast = sccun;
}

void finalize_stats(long totalc)
{
    int i;
    double a,prob[256];
    
    chisq=0;
    ent=0;
      
    double cexp = (double) totalc / 256.0;
    
    /* Compute average */
    average = aval / totalc;
    
    /* Compute chi square */
    for (i = 0; i < 256; i++) {
       prob[i] = (double) (ccount[i]) / totalc;
       a = ccount[i] - cexp;
       chisq = chisq + (a * a) / cexp;
    }

    /* Compute entropy */
    for (i = 0; i < 256; i++) {
       if (prob[i] > 0.0) {
	      ent += prob[i] * log10(1 / prob[i]) * log2of10;
       }
    }

    /* Find chi square probability */
	chip = sqrt(2.0 * chisq) - sqrt(2.0 *  255.0 - 1.0);
	a = fabs(chip);
	for (i = 9; i >= 0; i--) {
	   if (chsqt1[i] < a) break;
	}
	chip = (chip >= 0.0) ? chsqt0[i] : 1.0 - chsqt0[i];

    /* finish serial coefficient computation */
    double xscct1 = scct1 + scclast * sccu0;
    double xscct2 = scct2 * scct2;
    scc = totalc * scct3 - xscct2;
    if (scc == 0.0) {
       scc = -100000;
    } else {
       scc = (totalc * xscct1 - xscct2) / scc;
    }
}
    
void usage()
{
    fprintf(stderr,"rngfeed: reads entropy from a file or a T12RNG hardware");
    fprintf(stderr," number\n         generator, optionally tests it and ");
    fprintf(stderr,"dumps it to the Linux RNG pool.\n");
    fprintf(stderr,"Version 0.1 - By Marco \"Kiko\" Carnut, ");
    fprintf(stderr,"<kiko /at/ tempest /dot/ com /dot/ br>,\n");
    fprintf(stderr,"              03-jul-2006\n\n");
    fprintf(stderr,"Usage: rngfeed [options] <source>\n\n");
    fprintf(stderr,"   source: file with random bytes or serial port with\n");
    fprintf(stderr,"           a T12RNG device attached ");
    fprintf(stderr,"(default: " DEFAULT_INPUT_FILE ")\n\n");
    fprintf(stderr,"Options:\n\n");
    fprintf(stderr," -a        (always) dump data to outfile even if ");
    fprintf(stderr,"they didn't pass the tests\n");
    fprintf(stderr," -h        this help text.\n");
    fprintf(stderr," -l <file> specifies the log device, default is stdout\n");
    fprintf(stderr," -o <file> outputs random data to the specified file\n");
    fprintf(stderr," -p <num>  pool size in bytes\n");
    fprintf(stderr," -r <dev>  dumps tested data to this random device ");
    fprintf(stderr,"(typically /dev/random)\n\n");
}

/* Alllocates 'size' bytes in the heap, returning a pointer to it if successful
   or aborting with a message specifying which source 'file' and 'line' the
   call came from. Never used directly; it's supposed to be used via the
   NEW macro. */

void *kalloc(int size, char *file, int line)
{
    void *p=malloc(size);
    if (p) return p;
    fprintf(stderr,"fatal: memory allocation failure at %s:%d, ",file,line);
    fprintf(stderr,"%d bytes requested\n",size);
    exit(1); return NULL;
}
