#!/usr/bin/perl
#
# [KSC4-PERL]---------------------------------------------------------------[]
#
#  T12TST.PL - Tests the T12RNG, version 0.0.1
#              By Marco "Kiko" Carnut <kiko@tempest.com.br> 2006-mar-26
#
#  Description: This program continually reads hex digits form the serial
#  ------------ ports, converting them to binary data and running several
#  statistic tests to evaluate the randomness of the data. The calculations
#  are performed in a cumulative fashion that is CPU-friendly, making the
#  program suitable for very long (several days or weeks) runs.
#
#  The random data is saved in a file for later re-analysis, perhaps with
#  other tools such as George Marsaglia's DieHard tests or NIST's STS. The
#  stats are also saved in a logfile for later inspection. Data files are
#  are flushed every time they are written, so you can interrupt the program
#  at any time without losing more than the last second of data.
#
#  The interpretation of the results is somewhat of an art. The following
#  should be interpreted as a rule of thumb:
#
#  - bias0 should be close to 50% and get slowly closer as abyte increases.
#    It's ok for it to be >=51% or <=49% when abyte is less than 1K.
#  - devavg should be close to zero and get slowly closer as abyte increases.
#    It is ok for abs(devavg)>1 when abytes is less than 1K.
#  - entropy should be close to 8 as abyte goes beyond a couple Ks.
#  - chip should spend most of its time between 25% and 75%. It's ok for it
#    to stray every now and then below 10% or above 90%. But if it does that
#    all the time, it's a bad thing. Going over 97.5% or lower than 2.5% is
#    very bad.
#  - sercorr should be close to zero and keep slowly getting closer.
#
#  Acknowledgements: the statistics have been based on John Walker's "ent"
#  ----------------- program, with minor adaptations. 
#
#  Portability: Requires the Device::SerialPort module. Earlier versions
#  ------------ of Perl do not come by default with the Time::HiRes module,
#  so you may need to install it manually. Tested with Perl 5.8.7 under Cygwin
#  1.5.18 and Linux 2.4.29, but should work on any Unix-like system with a
#  reasonably modern Perl.
#
# []------------------------------------------------------------------------[]

use strict;

use Device::SerialPort qw( :PARAM :STAT 0.07 );
use Time::HiRes;

Device::SerialPort->set_test_mode_active(1);

my $retries=3;
my $device="/dev/ttyS1";

my @n1s;
init_n1s();

my @ccount;
my $ent;
my $chisq;
my $chip;
my ($scc, $sccun, $sccu0, $scclast, $scct1, $scct2, $scct3, $sccfirst);

my @chsqt0 = ( 0.5   , 0.25  , 0.1   , 0.05  , 0.025 ,
               0.01  , 0.005 , 0.001 , 0.0005, 0.0001 
);

my @chsqt1 = ( 0.0000, 0.6745, 1.2816, 1.6449, 1.9600,
               2.3263, 2.5758, 3.0902, 3.2905, 3.7190
);

my $log2of10=1.0/log(2);

my $logfile="t12test.log";
my $rawfiletmpl="t12test-%02d.dat";
my $rawfile="t12test-2.dat";

my $infile;
#$infile="/dev/urandom";

for (;;) {
    my $port;
    
    if ($infile) {
        open(INFH,$infile) || die "Can't open $infile for reading";
        print STDERR "Reading from $infile\n";
    } else {
        $port=Device::SerialPort->new($device,1);
        print STDERR "Reading from $device\n";
    }

    if ($port) {
        $port->databits(8);
        $port->baudrate(1200);
        $port->parity("none");
        $port->stopbits(1);
        $port->handshake("none");
        $port->read_char_time(0);     # don't wait for each character
        $port->read_const_time(1000); # 1 second per unfulfilled "read" call
    }

    my $timeout=5;
  
    my ($an0s,$an1s,$anbits,$abytes,$aval);
    my ($prev,$lines);

    reset_stats();
    
    my $t0=Time::HiRes::time();
    while ($timeout>0) {
        my ($count,$saw);
        if ($infile) {
            $count=125;
            read(INFH,$saw,$count);
            #print STDERR unpack("H*",$saw); exit;
            #sleep(1);
        } else {
            ($count,$saw)=$port->read(255); # will read _up to_ 255 chars
        }
        if ($count > 0) {
            my ($str,$l);
            my $now=Time::HiRes::time();
            if ($infile) {
                $str=$saw;
                $l=length($saw);
            } else {
                $saw =~ s/\r\nT.*?\r\n//o;
                $saw =~ s/[^0-9A-F]//og;
                $saw = $prev . $saw;
                $l=length($saw);
                if ($l%2) {
                    $prev=substr($saw,$l-1);
                    $saw =substr($saw,0,$l-1);
                } else {
                    $prev="";
                }
                $l=length($saw);
                $str=pack("H*",$saw);
            }
            my $bytes=length($str);
            #if ($l != 2*$bytes) {
            #     print STDERR "somethin's screwy: l=$l, bytes=$bytes!\n";
            #     print STDERR "saw=$saw\n";
            #}
            if (open(FH,">>$rawfile")) {
                print FH $str;
                close(FH);
            }
            my $n1s=n1s($str);
            my $nbits=$bytes*8;
            my $n0s=$nbits-$n1s;
            
            # Update accumulators for statistics
            $an0s   += $n0s;
            $an1s   += $n1s;
            $anbits += $nbits;
            $abytes += $bytes;
            my $bean=0;
            for (split(//,$str)) {
                my $c=ord($_);
                $aval+=$c;
                $ccount[$c]++;
                $sccun = $c;
                if ($sccfirst) {
                    $sccfirst = 0;
                    $scclast = 0;
                    $sccu0 = $sccun;
                } else {
                    $scct1 = $scct1 + $scclast * $sccun;
                }
                $scct2 = $scct2 + $sccun;
                $scct3 = $scct3 + ($sccun * $sccun);
                $scclast = $sccun;
            }
             
            # Finish stats so have partial results thus far
            update_stats($abytes);
            
            my $msg;
            $msg="rdb -Bps -abytes -bias0 -devavg entropy -dchisq ".
                "--chip --sercorr\n"
                unless $lines++%10;
            $msg.=sprintf "%3d ",$bytes;
            $msg.=sprintf "%4.1f ", $abytes/($now-$t0);
            $msg.=sprintf "%7d ",$abytes;
            $msg.=sprintf "%5.2f%% ",100*$an0s/$anbits;
            $msg.=sprintf "%+6.4f ",$aval/$abytes-127.5;
            $msg.=sprintf "%7.5f ", $ent;
            $msg.=sprintf "%+7.2f ", $chisq-255.0;
            $msg.=sprintf "%5.2f%% ", $chip*100;
            $msg.=sprintf "%+8.6f",$scc;
            $msg.="\n";
            out($msg);
            if ($infile && eof(INFH)) {
                $timeout=-1; last;
            }
       } else {
           $timeout--;
       }
    }
    
    last if $timeout<0;
    
    if ($timeout==0) {
        if (--$retries) {
            print STDERR "timed out, closing... ";
            undef $port;
            print STDERR "sleeping for 5 seconds... ";
            sleep(5);
            print STDERR "retrying...\n";
        } else {
            print STDERR "still nothing on the line, giving up...\n";
            exit 1;
        }
    }
}

exit;

sub init_n1s
{
    for (0..255) {
        my $c=chr($_);
        my $ones=grep(/1/,split(//,unpack("B8",$c)));
        $n1s[$_]=$ones;
	}
}

sub n1s
{
    my ($str) = @_;
    my $n1s=0;
    for (split(//,$str)) {
        $n1s+=$n1s[ord($_)];
    }
    return $n1s;
}

sub reset_stats
{
    for (0..255) { $ccount[$_]=0; } # clear bins
    $chisq=$ent=$chip=0;            # Clear entropy and chi square stats
    $sccfirst = 1;	                # Mark first time for serial correlation
    $scct1 = $scct2 = $scct3 = 0.0; # Clear serial correlation terms
    $scc=$sccun=$sccu0=$scclast=0;
}

sub update_stats
{
    my ($totalc)=@_;
    my ($i,$a);
    my @prob;
    
    $chisq=0;
    $ent=0;
      
    my $cexp = $totalc / 256.0;
    my $datasum=0;
    
    # compute chi square
    for ($i = 0; $i < 256; $i++) {
       $prob[$i] = $ccount[$i] / $totalc;
       $a = $ccount[$i] - $cexp;
       $chisq = $chisq + ($a * $a) / $cexp;
       $datasum += $i * $ccount[$i];
    }

    # Compute entropy
    for ($i = 0; $i < 256; $i++) {
       if ($prob[$i] > 0.0) {
	      $ent += $prob[$i] * log(1 / $prob[$i]) * $log2of10;
       }
    }

    # find chi square probability
	$chip = sqrt(2.0 * $chisq) - sqrt(2.0 *  255.0 - 1.0);
	$a = abs($chip);
	for ($i = 9; $i >= 0; $i--) {
	   last if $chsqt1[$i] < $a;
	}
	$chip = ($chip >= 0.0) ? $chsqt0[$i] : 1.0 - $chsqt0[$i];

    # finish serial coefficient computation   
    my $xscct1 = $scct1 + $scclast * $sccu0;
    my $xscct2 = $scct2 * $scct2;
    $scc = $totalc * $scct3 - $xscct2;
    if ($scc == 0.0) {
       $scc = -100000;
    } else {
       $scc = ($totalc * $xscct1 - $xscct2) / $scc;
    }
}

sub out
{
    my ($msg)=@_;
    print STDERR $msg;
    if ($logfile && open(FH,">>$logfile")) {
        print FH $msg;
        close(FH);
    }
}
